import math
import numbers

from six import integer_types
from six import string_types

from hackle import model


def is_non_empty_string(input_id_key):
    if input_id_key and isinstance(input_id_key, string_types):
        return True

    return False


def is_non_zero_and_empty_int(input_id_key):
    if input_id_key and isinstance(input_id_key, integer_types) and input_id_key > 0:
        return True

    return False


def is_number(input_id_key):
    if input_id_key and isinstance(input_id_key, numbers.Number):
        return True

    return False


def is_event_value_valid(attribute_value):
    if attribute_value is None:
        return True

    if isinstance(attribute_value, (numbers.Integral, float)):
        return is_finite_number(attribute_value)

    return False


def is_finite_number(value):
    if not isinstance(value, (numbers.Integral, float)):
        # numbers.Integral instead of int to accommodate long integer in python 2
        return False

    if isinstance(value, bool):
        # bool is a subclass of int
        return False

    if isinstance(value, float):
        if math.isnan(value) or math.isinf(value):
            return False

    if abs(value) > (2 ** 53):
        return False

    return True


def is_valid_user(user):
    if user is None:
        return False

    if user and not isinstance(user, model.User):
        return False

    if user and user.id is None:
        return False

    if user and not isinstance(user.id, string_types):
        return False

    return True


def is_valid_event(event):
    if event is None:
        return False

    if event and not isinstance(event, model.Event):
        return False

    if event and event.key is None:
        return False

    if event and not isinstance(event.key, string_types):
        return False

    if event and event.value and not is_event_value_valid(event.value):
        return False

    return True



import logging


class Default(object):
    WORKSPACE_URL = 'https://sdk.hackle.io/api/v1/workspaces'
    EVENT_URL = 'https://event.hackle.io/api/v1/events'
    SDK_KEY_HEADER = 'X-HACKLE-SDK-KEY'
    SDK_NAME_HEADER = 'X-HACKLE-SDK-NAME'
    SDK_VERSION_HEADER = 'X-HACKLE-SDK-VERSION'
    CONTENT_TYPE_HEADER = 'Content-Type'
    SDK_CONTENT_TYPE = 'application/json'
    DEFAULT_POLL_INTERVAL = 10
    DEFAULT_BLOCKING_TIMEOUT = 5
    REQUEST_TIMEOUT = 60


class HTTPVerbs(object):
    GET = 'GET'
    POST = 'POST'


class DecisionType(object):
    NATURAL = 'NATURAL'
    FORCED = 'FORCED'


class EventType(object):
    EXPOSURE = 'EXPOSURE'
    TRACK = 'TRACK'


class LogLevels(object):
    NOTSET = logging.NOTSET
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL



import os

from setuptools import setup
from setuptools import find_packages

here = os.path.join(os.path.dirname(__file__))

__version__ = None
with open(os.path.join(here, 'hackle', 'version.py')) as _file:
    exec(_file.read())

with open(os.path.join(here, 'requirements', 'sdk.txt')) as _file:
    REQUIREMENTS = _file.read().splitlines()

with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='hackle-sdk',
    version=__version__,
    author='Hackle',
    author_email='developers@hackle.io',
    packages=find_packages(exclude=['docs', 'tests']),
    url='https://hackle.io',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=REQUIREMENTS,
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ]
)



from collections import namedtuple

from . import bucketer
from . import entities
from .commons import enums

Decision = namedtuple('Decision', 'experiment variation type')


class Decider(object):
    def __init__(self, logger):
        self.bucketer = bucketer.Bucketer()
        self.logger = logger

    def decide(self, experiment, user):
        if isinstance(experiment, entities.CompletedExperiment):
            return Decision(None, experiment.winnerVariationKey, enums.DecisionType.FORCED)

        if isinstance(experiment, entities.RunningExperiment):
            return self.__decide(experiment, user)

    def __decide(self, experiment, user):
        variation = experiment.userOverrides.get(user.id)
        if variation:
            return Decision(None, variation.key, enums.DecisionType.FORCED)

        slot = self.bucketer.bucketing(experiment.bucket, user.id)
        if not slot:
            return None

        variation = experiment.variations[slot.variationId]

        if not variation:
            return None

        if variation.isDropped:
            return None

        return Decision(experiment, variation, enums.DecisionType.NATURAL)
